#ifndef _SOUND_H

using namespace std;

#include "tester.h"
#include "bipointer.h"
#include "3dmath.h"

class SOUNDINFO
{
friend class SOUNDBUFFER;
friend class SOUNDSOURCE;
	
private:
	unsigned int samples, frequency, bytespersample, channels;
	
public:
	SOUNDINFO() : samples(0), frequency(0), bytespersample(0), channels(0) { }
	void Set(const int newsamp, const int newfreq, const int newchan, const int newbytes) {samples = newsamp;frequency = newfreq;channels=newchan;bytespersample=newbytes;}
	void DebugPrint() const;
	const int GetSamples() const {return samples;}
	const int GetFrequency() const {return frequency;}
	const int GetChannels() const;
	const int GetBytesPerSample() const {return bytespersample;}
	
	bool operator==(const SOUNDINFO & other) const {return (samples == other.samples && frequency == other.frequency && channels == other.channels && bytespersample == other.bytespersample);}
};

class SOUNDBUFFER
{
friend class SOUNDSOURCE;
private:
	//SOUNDINFO original_info;
	SOUNDINFO info;
	unsigned int size;
	bool loaded;
	char * sound_buffer;
	string name;
	void LoadWAV(const string & filename, const SOUNDINFO & sound_device_info);
	void LoadOGG(const string & filename, const SOUNDINFO & sound_device_info);

public:
	SOUNDBUFFER() : loaded(false),sound_buffer(NULL) {}
	~SOUNDBUFFER() {Unload();}
	
	void Load(const string & filename, const SOUNDINFO & sound_device_info);
	void Unload() {if (loaded && sound_buffer != NULL) delete [] sound_buffer;sound_buffer = NULL;}
	const SOUNDINFO & GetSoundInfo() const {return info;}
	//const SOUNDINFO & GetOriginalInfo() {return original_info;}
	//const unsigned int GetSize() {return size;}
	//void ConvertTo(const SOUNDINFO & to_info);
	const int GetSample16bit(const unsigned int channel, const unsigned int position);
	const string & GetName() const {return name;}
};

#define MAX_FILTER_ORDER 5

class SOUNDFILTER
{
private:
	int order;
	float xc[MAX_FILTER_ORDER+1];
	float yc[MAX_FILTER_ORDER+1];
	float statex[2][MAX_FILTER_ORDER+1];
	float statey[2][MAX_FILTER_ORDER+1];

public:
	SOUNDFILTER() : order(0) {ClearState();}
	
	void ClearState() {for (int c = 0; c < 2; c++) for (int i = 0; i < MAX_FILTER_ORDER; i++) {statex[c][i] = 0.0;statey[c][i] = 0.0;}}
	void SetFilter(const int neworder, const float * xcoeff, const float * ycoeff); //the coefficients are arrays like xcoeff[neworder+1], note ycoeff[0] is ignored
	void Filter(int * chan1, int * chan2, const int len);
	void SetFilterOrder1(float xc0, float xc1, float yc1); //yc0 is ignored
	void SetFilterOrder0(float xc0); //yc0 is ignored
};

class SOUNDSOURCE : public BIPOINTER_TARGET
{
private:
	unsigned int sample_pos;
	unsigned int last_sample_pos;
	float sample_pos_remainder;
	int playing;
	int loop;
	bool autodelete;
	float gain;
	float pitch;
	float last_pitch;
	VERTEX position;
	VERTEX velocity;

	float computed_gain1;
	float computed_gain2;
	
	float last_computed_gain1;
	float last_computed_gain2;

	bool effects3d;

	float relative_gain;

	SOUNDBUFFER * buffer; //used for speed

	list <SOUNDFILTER> filters;

public:
	SOUNDSOURCE() : sample_pos(0),last_sample_pos(0),sample_pos_remainder(0.0f),playing(0),loop(false),autodelete(false),gain(1.0),pitch(1.0),last_pitch(1.0),computed_gain1(1.0),computed_gain2(1.0),last_computed_gain1(0.0),last_computed_gain2(0.0),effects3d(true),relative_gain(1.0),buffer(NULL) {}
	void SetBuffer(SOUNDBUFFER & newbuf) {/*buffer.SetPtr(newbuf);*/buffer=&newbuf;}
	void SampleAndAdvanceWithPitch16bit(int * chan1, int * chan2, int len);
	void IncrementWithPitch(int num);
	void Increment(int num);
	void SampleAndAdvance16bit(int * chan1, int * chan2, int len);
	void Sample16bit(unsigned int peekoffset, int & chan1, int & chan2);
	void Advance(unsigned int offset);
	void SeekToSample(const unsigned int newpos) {sample_pos = newpos;}
	void SetAutoDelete(const bool newauto) {autodelete = newauto;}
	bool GetAutoDelete() const {return autodelete;}
	void SetGain(const float newgain) {gain = newgain;}
	void SetPitch(const float newpitch) {pitch = newpitch;}
	void SetGainSmooth(const float newgain, const float dt);
	void SetPitchSmooth(const float newpitch, const float dt);
	void SetPosition(const VERTEX newpos) {position.Set(newpos);}
	const VERTEX & GetPosition() {return position;}
	void SetVelocity(const VERTEX newvel) {velocity.Set(newvel);}
	const VERTEX & GetVelocity() {return velocity;}
	void Set3DEffects(bool new3d) {effects3d = new3d;}
	const bool Get3DEffects() const {return effects3d;}
	void SetComputationResults(float cpg1, float cpg2) //{last_computed_gain1 = computed_gain1;last_computed_gain2 = computed_gain2;computed_gain1 = cpg1;computed_gain2 = cpg2;last_sample_pos = sample_pos;}
		{computed_gain1 = cpg1;computed_gain2 = cpg2;}
	const float GetGain() const {return gain;}
	const float ComputedGain(const int channel) const {if (channel == 1) return computed_gain1; else return computed_gain2;}
	void SetRelativeGain(const float relgain) {relative_gain = relgain;}
	const float GetRelativeGain() {return relative_gain;}
	
	void SetLoop(const bool newloop) {if (newloop) loop = 1; else loop = 0;}
	void Reset() {SeekToSample(0);}
	void Stop() {playing = 0;Reset();}
	void Pause() {playing = 0;}
	void Play() {playing = 1;}
	//const bool Audible() const {return (playing == 1 && (computed_gain1 > 0.0 || computed_gain2 > 0.0));}
	const bool Audible() const;
	const string GetName() const {if (buffer == NULL) return "NULL"; else return buffer->GetName();}
	
	const SOUNDBUFFER & GetSoundBuffer() const {return *buffer;}
	SOUNDFILTER & AddFilter() {{SOUNDFILTER newfilt;filters.push_back(newfilt);}return filters.back();}
	SOUNDFILTER & GetFilter(int num);
	const int NumFilters() const {return filters.size();}
	void ClearFilters() {filters.clear();}
};

class SOUND
{
private:
	TESTER mytest;

	bool loaded;
	bool initdone;
	bool paused;

	SOUNDINFO deviceinfo;

	map <string, SOUNDBUFFER> buffermap;
	list <SOUNDSOURCE> sourcelist;

	void LoadBuffer(const string & buffername);
	void DetermineActiveSources(list <SOUNDSOURCE *> & active_sourcelist, list <SOUNDSOURCE *> & inaudible_sourcelist);

	float gain_estimate;
	SOUNDFILTER volume_filter;
	bool disable;

	VERTEX lpos, lvel;
	QUATERNION lrot;

public:
	SOUND() : loaded(false),initdone(false),paused(true),gain_estimate(1.0),disable(false) {volume_filter.SetFilterOrder0(1.0);}
	~SOUND();
	
	const TESTER & Test();
	
	void Init();
	void Load();
	void Callback16bitstereo(void *unused, Uint8 *stream, int len);
	void Pause(const bool pause_on);
	SOUNDSOURCE & NewSource(const string & buffername);
	void Tick();
	void DeleteSource(SOUNDSOURCE * todel);
	void NewSourcePlayOnce(const string & buffername);
	void Compute3DEffects(list <SOUNDSOURCE *> & sources, const VERTEX & listener_pos, const QUATERNION & listener_rot);
	void Compute3DEffects(list <SOUNDSOURCE *> & sources) {Compute3DEffects(sources, lpos, lrot);}
	void UpdateSettings();
	void Unload() {sourcelist.clear();buffermap.clear();}
	void Reload() {Unload();Load();}
	void SetMasterVolume(float newvol) {volume_filter.SetFilterOrder0(newvol);}
	void DisableAllSound() {disable=true;}
	void SetListener(const VERTEX & npos, const QUATERNION & nrot, const VERTEX & nvel) {lpos = npos; lrot = nrot; lvel = nvel;}
	const bool Enabled() const {return !disable;}
};

#define _SOUND_H
#endif
