/***************************************************************************
 *            textures.h
 *
 *  Thu Dec  1 17:46:28 2005
 *  Copyright  2005  Joe Venzon
 *  joe@venzon.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _TEXTURES_H

#include <list>

#include "exception.h"
#include "bipointer.h"

using namespace std;
using namespace VDrift;

class TEXTUREINFO
{
private:
	string name;
	bool mipmap;
	bool cube;
	bool verticalcross;
	bool normalmap;
	SDL_Surface * surface;
	
public:
	TEXTUREINFO() {SetDefaults();}
	void SetDefaults() {mipmap = true;cube=false;verticalcross=false;normalmap=false;surface=NULL;}
	const string GetName() const {return name;}
	const bool GetMipMap() const {return mipmap;}
	const bool GetCube() const {return cube;}
	const bool GetVerticalCross() const {return cube && verticalcross;}
	SDL_Surface * GetSurface() {return surface;}
	const bool NormalMap() const {return normalmap;}
	//void CopyFrom(const TEXTUREINFO & other) {name = other.GetName(); mipmap = other.GetMipMap();cube=other.GetCube();verticalcross=other.GetVerticalCross();}
	void CopyFrom(const TEXTUREINFO & other) {*this = other;}
	void SetName(const string & newname) {name = newname;}
	void SetCube(const bool newcube, const bool newvertcross) {cube = newcube;verticalcross=newvertcross;}
	void SetNormalMap(const bool newnorm) {normalmap = newnorm;}
	void SetMipMap(const bool newmipmap) {mipmap = newmipmap;}
	void SetSurface(SDL_Surface * newsurface) {surface = newsurface;}
};

class TEXTURE_SDLGL : public BIPOINTER_TARGET
{
private:
	TEXTUREINFO texture_info;
	GLuint tex_id;
	bool loaded;
	unsigned short int w, h;
	int references;
	bool alphachannel;
	void LoadCube();
	void LoadCubeVerticalCross();
	
public:
	TEXTURE_SDLGL() { loaded = false; tex_id = 0; w = h = 0; references = 0; alphachannel = false;}
	~TEXTURE_SDLGL() { Unload(); }

	void SetInfo(const TEXTUREINFO & texinfo) {texture_info.CopyFrom(texinfo);}
	void Load();
	void Unload();
	const unsigned short int GetW() const {return w;}
	const unsigned short int GetH() const {return h;}
	const bool IsEqualTo(const TEXTURE_SDLGL & othertex) const {return IsEqualTo(othertex.GetTextureInfo());}
	const bool IsEqualTo(const TEXTUREINFO & texinfo) const;
	const TEXTUREINFO & GetTextureInfo() const {return texture_info;}
	void Activate() const {Activate(false);}
	void Activate(bool blur_override) const;
	void IncrementReferences() {references++;}
	void DecrementReferences();
	const int NumReferences() const {return references;}
};

class TEXTURE_HANDLE : public BIPOINTER_TARGET
{
private:
	BIPOINTER <TEXTURE_SDLGL> tex;
	
public:
	TEXTURE_HANDLE() {}
	TEXTURE_HANDLE(const TEXTURE_HANDLE & other);
	~TEXTURE_HANDLE();
	
	const bool Load( const TEXTUREINFO & texinfo );
	const bool Load( string new_tex_file ) { return Load( new_tex_file, true ); }
	const bool Load( string new_tex_file, bool mipmap ) { int w, h; return Load( new_tex_file, mipmap, w, h ); }
	const bool Load( string new_tex_file, bool mipmap, int &w, int &h ) {return Load( new_tex_file, NULL, mipmap, false, w, h );}
	const bool Load( string new_tex_file, SDL_Surface* surface, bool mipmap, bool cube, int &w, int &h );
	const bool Load( string new_tex_file, SDL_Surface* surface ) { int w, h; return Load( new_tex_file, surface, true, false, w, h ); }
	const bool Load( string new_tex_file, SDL_Surface* surface, bool mipmap ) { int w, h; return Load( new_tex_file, surface, mipmap, false, w, h ); }
	const bool Load( string new_tex_file, bool mipmap, bool cube, int &w, int &h ) { return Load( new_tex_file, NULL, mipmap, cube, w, h); }
	
	void Unload();
	void Activate(bool blur_override) const;
	void Activate() const {Activate(false);}
	
	const bool Loaded() const {return (tex.PtrValid());}
	
	const TEXTURE_SDLGL & GetTexture() const {return *tex.GetPtr();}
	
	TEXTURE_HANDLE & CopyFrom(const TEXTURE_HANDLE & other);
	TEXTURE_HANDLE & operator=(const TEXTURE_HANDLE & other) {return CopyFrom(other);}
};

class TEXTURES
{
private:
	list <TEXTURE_SDLGL> textures;

	TEXTURE_SDLGL * FindTexture (const TEXTUREINFO & texinfo);
	
public:
	TEXTURE_SDLGL * LoadTexture(const TEXTUREINFO & texinfo ) {int w,h;return LoadTexture(texinfo,w,h);}
	TEXTURE_SDLGL * LoadTexture(const TEXTUREINFO & texinfo, int &w, int &h );
	void UnloadTexture(const TEXTURE_SDLGL & textodel);
	
	void ReloadAll();
};

#define _TEXTURES_H
#endif /* _TEXTURES_H */
