/* vim: set noexpandtab shiftwidth=8 cino=:
 ***************************************************************************
 *            logging.h
 *
 *  Copyright  2005  Matthew A. Nicholson
 *  matt@matt-land.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "logging.h"
#include <iostream>

using namespace VDrift;
using namespace std;

//using std::tr1::shared_ptr;

/** Set the current logging level to LOG_CRITICAL.
 */
LOGGER& VDrift::critical(LOGGER& stream)
{
	return stream.SetLevel(LOG_CRITICAL);
}

/** Set the current logging level to LOG_ERROR.
 */
LOGGER& VDrift::error(LOGGER& stream)
{
	return stream.SetLevel(LOG_ERROR);
}

/** Set the current logging level to LOG_WARN.
 */
LOGGER& VDrift::warn(LOGGER& stream)
{
	return stream.SetLevel(LOG_WARN);
}

/** Set the current logging level to LOG_INFO.
 */
LOGGER& VDrift::info(LOGGER& stream)
{
	return stream.SetLevel(LOG_INFO);
}

/** Set the current logging level to LOG_DEBUG.
 */
LOGGER& VDrift::debug(LOGGER& stream)
{
	return stream.SetLevel(LOG_DEBUG);
}

/** Get the specified logger.
 * @param log The name of the logger to get.
 *
 * As long as the same name is specified, the same LOGGER wil be returned.
 *
 * @return A reference to the indicated LOGGER.
 */
LOGGER& LOGGING::GetLogger(const std::string& logger)
{
	static LOGGING instance;
	return instance.loggers[logger];
}

/// Empty constructor.
LOGGING::LOGGING()
{
}

/** Construct a logger with the specified level.
 * @param default_level The inital default level of this logger.
 */
LOGGER::LOGGER(unsigned int default_level) : ostream(), buffer(default_level)
{
	this->init(&buffer);
}

/** Construct a logger from another logger.
 * @param logger The logger to copy.
 */
LOGGER::LOGGER(const LOGGER& logger) : ostream(), buffer(logger.buffer.level)
{
	this->init(&buffer);
}

/** Add a log to this logger.
 * @param log The log to add.
 *
 * This function adds the indicated LOG to this logger.  The log to be added is
 * expected as a LOGGER::log_t type which is an std::tr1::shared_ptr< LOG >.
 * Changed to a LOG*....
 */
void LOGGER::AddLog(LOGGER::log_t log)
{
	buffer.logs.push_back(log);
}

/** Add a stream to this logger.
 * @param stream A pointer to the stream to add.
 * @param level The level for the new stream.
 *
 * This function is a helper function that creates a STREAMLOG object from the
 * specified stream and adds it to the LOGGER.
 *
 * @note The stream pointed to will not be deleted automatically.  If this
 * functionality is desired, use a shared_ptr.
 * Don't worry, I'll fix it. ;)
 */
void LOGGER::AddLog(std::ostream* stream, unsigned int level)
{
	buffer.logs.push_back((LOG*)(new STREAMLOG< std::ostream* >(stream, level)));
}

/** Add a stream to this logger.
 * @param stream A shared_ptr to the stream to add.
 * @param level The level for the new stream.
 *
 * This function is a helper function that creates a STREAMLOG object from the
 * specified stream and adds it to the LOGGER.
 * 
 * Removed to shake loose dependency on shared ptr
 */
/*
void LOGGER::AddLog(shared_ptr<std::ostream> stream, unsigned int level)
{
	buffer.logs.push_back(shared_ptr< LOG >(new STREAMLOG< shared_ptr< std::ostream > >(stream, level)));
}
*/

/** Remove a stream from this logger.
 * @param log The log to remove.
 */
void LOGGER::RemoveLog(LOGGER::log_t log)
{
	buffer.logs.remove(log);
}


/** Set the logging threshold for this logger.
 * @param new_level The new threshold.
 * 
 * This function sets the logging threshold for this logger.  LOG objects that
 * have a logging level greater than the the threshold will not recieve any
 * messages.  For example a LOG set to LOG_CRITICAL will not get messages if
 * this LOGGER is set at LOG_DEBUG.
 *
 * @note LOG_NONE can be used to turn of log output.
 */
LOGGER& LOGGER::SetLevel(unsigned int level)
{
	buffer.level = level;
	return *this;
}

/** Overloaded inserter operator (<<) to allow I/O manip type functions with
 * LOGGER streams.
 */
LOGGER& LOGGER::operator << (LOGGER& (*func)(LOGGER&))
{
	return func(*this);
}

/** Construct a STREAMLOG pointing to the indicated stream with the 
 * indicated level.
 * @param log_stream The stream to point to.
 * @param level The level of the stream.
 */
/*
template<typename ptr_t>
STREAMLOG<ptr_t>::STREAMLOG(ptr_t log_stream, unsigned int level) : LOG(level), stream(log_stream)
{
}
*/

/** Construct a LOGBUF.
 * @internal
 * @param default_level The level of the logbuf.
 */
LOGBUF::LOGBUF(unsigned int default_level) : std::streambuf(), level(default_level)
{
}

/** A reimplimentation of std::streambuf::overflow().
 * @param c The character to output.
 * \return The given character.
 */
std::streambuf::int_type LOGBUF::overflow(std::streambuf::int_type c)
{
	if(level != LOG_NONE)
	{
		for(std::list< log_t >::iterator i = logs.begin(); i != logs.end(); ++i)
		{
			if((*i)->GetLevel() >= level)
				(*i)->put(c);
				
		}

	}
	return c;
}

/** An implimentation of std::streambuf::sync()
 * \return -1 on failure
 */
int LOGBUF::sync()
{
	for(std::list< log_t >::iterator i = logs.begin(); i != logs.end(); ++i)
	{
		(*i)->flush();
	}
	return 0;
}


