//  Gl_Car.cc - a car that handles graphics, sound and input devices.
//
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/body/Gl_Car.h>
//#include <vamos/geometry/Ac3d.h>
#include <vamos/geometry/Conversions.h>
#include <vamos/body/Fuel_Tank.h>
#include <vamos/body/Gauge.h>
#include <vamos/body/Wheel.h>

#include "singleton.h"
#include "game.h"
extern GAME game;

using namespace Vamos_Geometry;

//* Struct Rear_View_Mirror

//** Constructor
Vamos_Body::
Rear_View_Mirror::Rear_View_Mirror (const Three_Vector& position,
									double width, double height,
									double direction,
									double field,
									double near_plane, double far_plane,
									std::string mask_file) :
  m_position (position),
  m_width (width),
  m_height (height),
  m_direction (direction),
  m_field (field),
  m_near_plane (near_plane),
  m_far_plane (far_plane),
  mp_mask (new Gl_Texture_Image (mask_file, false, false))
{
}

//** Destructor
Vamos_Body::
Rear_View_Mirror::~Rear_View_Mirror ()
{
  delete mp_mask;
}

void Vamos_Body::
Rear_View_Mirror::activate_viewport ()
{
  glViewport (m_viewport.x, m_viewport.y, m_viewport.width, m_viewport.height);
  glScissor (m_viewport.x, m_viewport.y, m_viewport.width, m_viewport.height);
}

Three_Vector Vamos_Body::
Rear_View_Mirror::get_center () const
{
  return Three_Vector (m_position [0], 
					   m_position [1] - m_width / 2.0,
					   m_position [2] + m_height / 2.0);
}

void Vamos_Body::
Rear_View_Mirror::transform_view () const
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  // Reflect the x-axis.
  glScaled (-1.0, 1.0, 1.0);
  gluPerspective (m_field, m_viewport.aspect (), m_near_plane, m_far_plane);
}

void Vamos_Body::
Rear_View_Mirror::set_view ()
{
  activate_viewport ();
  glClear (GL_DEPTH_BUFFER_BIT);
  transform_view ();
}

//* Class Gl_Car

extern bool verbose_output;

//** Constructor
Vamos_Body::
Gl_Car::Gl_Car (const Three_Vector& pos)
  : Car (pos),
	mp_dashboard (0),
	cur_paintjob(0),
	num_paintjobs(0)
{
}

//** Destructor
Vamos_Body::
Gl_Car::~Gl_Car ()
{
  delete mp_dashboard;

  for (std::vector <Rear_View_Mirror*>::iterator it = m_mirrors.begin ();
	   it != m_mirrors.end ();
	   it++)
	{
	  delete *it;
	}
	
	for (map <float, BIPOINTER <SOUNDSOURCE> >::iterator i = engine_samples.begin(); i != engine_samples.end(); i++)
	{
		if (i->second.PtrValid())
			game.sound.DeleteSource(i->second.GetPtr());
	}
	
	for (int i = 0; i < 4 ; i++)
		if (tire_source[i].PtrValid())
			game.sound.DeleteSource(tire_source[i].GetPtr());
	
	if (collision_drawable.PtrValid())
		collision_node.GetRef().Delete(collision_drawable.GetPtr());
	if (collision_node.PtrValid())
		game.graphics.GetScene().GetRoot().Delete(collision_node.GetPtr());
	
	if (cm_node.PtrValid())
	{
		if (exterior_drawable.PtrValid())
			origin_node.GetRef().Delete(exterior_drawable.GetPtr());
		if (interior_drawable.PtrValid())
			origin_node.GetRef().Delete(interior_drawable.GetPtr());
		if (glass_drawable.PtrValid())
			origin_node.GetRef().Delete(glass_drawable.GetPtr());
		cm_node.GetRef().Delete(origin_node.GetPtr());
		cm_node.GetRef().GetParent()->Delete(cm_node.GetPtr());
	}
}

#define LOADFLOATING(floatmodel, floatfile) {string floatingfile = game.settings.GetFullDataPath("cars/" + m_car_file + "/" + floatfile + ".joe"); \
	if (game.utility.FileExists(floatingfile)) \
	{ \
		floatmodel.Load(floatingfile,  false); \
		char texfilename[1024]; \
		sprintf(texfilename, "cars/%s/textures/body%02i.joe",m_car_file.c_str(), cur_paintjob); \
		floatmodel.Texture(game.settings.GetFullDataPath(texfilename), 0); \
		floatmodel.AdditiveTexture(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/brake.joe"), 1); \
	}} \

void Vamos_Body::
Gl_Car::exterior_model (SCENENODE * parentnode)
{
	if (cm_node.PtrValid() || origin_node.PtrValid())
		throw EXCEPTION(__FILE__,__LINE__,"Tried to assign models to car twice!");
	
	//shadowtex.Load(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/shadow.png"), false);
	//joeglass.AdditiveTexture(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/brake-glass.joe"), 1);
	joeexterior.Load(game.settings.GetFullDataPath("cars/" + m_car_file + "/body.joe"));
	//joeexterior.AdditiveTexture(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/brake.joe"), 1);
		
	//LOADFLOATING(joefloating_front, "floating_front");
	//LOADFLOATING(joefloating_rear, "floating_rear");
	
	joecollision.Load(game.settings.GetFullDataPath("cars/" + m_car_file + "/collision.joe"));
	
	//check for paint jobs
	char texfile[1024];
	int count = 0;
	sprintf(texfile, "cars/%s/textures/body%02i.png", m_car_file.c_str(), count);
	while (count < 99 && game.utility.FileExists(game.settings.GetFullDataPath(texfile)))
	{
		count++;
		sprintf(texfile, "cars/%s/textures/body%02i.png", m_car_file.c_str(), count);
	}
	num_paintjobs = count;
	
	if (parentnode != NULL)
	{
		cm_node.SetPtr(parentnode->AddNode());
		origin_node.SetPtr(cm_node.GetRef().AddNode());
		
		const bool collide_debug(false);
		
		if (collide_debug)
			collision_node.SetPtr(game.graphics.GetScene().GetRoot().AddNode());
		
		//add exterior
		{
			DRAWABLE & dext = origin_node.GetRef().AddDrawable();
			exterior_drawable.SetPtr(dext);
			dext.SetToList(joeexterior.GetDrawList());
			dext.SetBlur(false);
			dext.SetMetallic(true);
			dext.SetPartialTransparency(true);
			//dext.SetDrawEnable(false);
			//cout << "Exterior set to " << joeexterior.GetDrawList() << endl;
		}
		
		//add glass
		if (game.utility.FileExists(game.settings.GetFullDataPath("cars/" + m_car_file + "/glass.joe")))
		{
			joeglass.Load(game.settings.GetFullDataPath("cars/" + m_car_file + "/glass.joe"));
			
			if (origin_node.PtrValid())
			{
				DRAWABLE & d = origin_node.GetRef().AddDrawable();
				glass_drawable.SetPtr(d);
				d.SetToList(joeglass.GetDrawList());
				d.SetBlur(false);
				d.SetMetallic(true);
				TEXTUREINFO texinfo;
				texinfo.SetName("cars/" + m_car_file + "/textures/glass.png");
				d.SetDiffuseMap(texinfo);
				d.SetPartialTransparency(true);
				if (game.utility.FileExists(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/glass-misc1.png")))
				{
					texinfo.SetName("cars/" + m_car_file + "/textures/glass-misc1.png");
					d.SetMiscMap1(texinfo);
				}
				else
				{
					stringstream msg;
					msg << "Car " << m_car_file << " has no gloss map for its glass; using a default constant shiny/reflective texture" << endl;
					game.WriteDebuggingData(msg.str());
					
					texinfo.SetName("textures/effects/white.png");
					d.SetMiscMap1(texinfo);
				}
			}
		}
		
		if (collision_node.PtrValid())
		{
			DRAWABLE & dcol = collision_node.GetRef().AddDrawable();
			collision_drawable.SetPtr(dcol);
			dcol.SetToList(joecollision.GetDrawList());
		}
	}
}

void Vamos_Body::
Gl_Car::interior_model (SCENENODE * parentnode)
{
  /*if (m_interior_list_id != 0)
	{
	  glDeleteLists (m_interior_list_id, 1);
	}

   Ac3d model (file, scale, translation, rotation);
   m_interior_list_id = model.build ();*/
	
	//joeinterior.Load(file);
	if (game.utility.FileExists(game.settings.GetFullDataPath("cars/" + m_car_file + "/interior.joe")))
	{
		joeinterior.Load(game.settings.GetFullDataPath("cars/" + m_car_file + "/interior.joe"));
		
		//add interior
		if (origin_node.PtrValid())
		{
			DRAWABLE & d = origin_node.GetRef().AddDrawable();
			interior_drawable.SetPtr(d);
			d.SetToList(joeinterior.GetDrawList());
			d.SetBlur(false);
			d.SetMetallic(true);
			TEXTUREINFO texinfo;
			texinfo.SetName("cars/" + m_car_file + "/textures/interior.png");
			d.SetDiffuseMap(texinfo);
			if (game.utility.FileExists(game.settings.GetFullDataPath("cars/" + m_car_file + "/textures/interior-misc1.png")))
			{
				texinfo.SetName("cars/" + m_car_file + "/textures/interior-misc1.png");
				d.SetMiscMap1(texinfo);
			}
		}
		else
			cout << "poo." << endl;
	}
}

void Vamos_Body::
Gl_Car::set_perspective (double aspect)
{
  gluPerspective (m_field_of_view, aspect, m_near_plane, m_far_plane);
}

void Vamos_Body::
Gl_Car::set_view (const Vamos_Geometry::Three_Vector& position,
				  double field_of_view,
				  double near_plane, double far_plane,
				  double pan_angle)
{
  m_driver_view = position;
  m_field_of_view = field_of_view;
  m_near_plane = near_plane;
  m_far_plane = far_plane;
  m_pan_angle = pan_angle;
}

void Vamos_Body::
Gl_Car::set_hood_view (const Vamos_Geometry::Three_Vector& position)
{
  m_hood_view = position;
}

void Vamos_Body::
Gl_Car::add_rear_view (const Vamos_Geometry::Three_Vector& position,
					   double width, double height,
					   double direction, double field,
					   double near_plane, double far_plane,
					   std::string mask_file)
{
  m_mirrors.push_back (new Rear_View_Mirror (position, width, height,
											 direction, field, 
											 near_plane, far_plane,
											 mask_file));
}


// Fill the stencil buffer for masking the rear-view mirrors.
void Vamos_Body::
Gl_Car::make_rear_view_mask (int window_width, int window_height)
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  glViewport (0, 0, window_width, window_height);
  glScissor (0, 0, window_width, window_height);

  glClearColor (0.0, 0.0, 0.0, 0.0);
  glClearStencil (0);
  glClear (GL_COLOR_BUFFER_BIT 
		   | GL_DEPTH_BUFFER_BIT
		   | GL_STENCIL_BUFFER_BIT);

  const double near_plane = 0.2;
  const double far_plane = 10.0;
  gluPerspective (field_of_view (), double (window_width)/window_height, 
				  near_plane, far_plane);
  view (0.0);
  glMatrixMode (GL_MODELVIEW);
  transform_body ();

  for (std::vector <Rear_View_Mirror*>::iterator it = m_mirrors.begin ();
	   it != m_mirrors.end ();
	   it++)
	{
	  (*it)->make_mask (window_width, window_height,
						m_driver_view, field_of_view ());
	}
}


// Put a mask for one rear-view mirror in the stencil buffer.
void Vamos_Body::
Rear_View_Mirror::make_mask (int window_width, int window_height,
							 const Three_Vector& driver_position, 
							 double driver_field_of_view)
{
  glDisable (GL_LIGHTING);
  set_viewport (window_width, window_height, 
				driver_position, driver_field_of_view);
  draw_mask_shape ();
  set_stencil (window_width, window_height);
  glEnable (GL_LIGHTING);
}


// Find the dimensions for a viewport that's just large enough to hold
// the mirror.
void Vamos_Body::
Rear_View_Mirror::set_viewport (int window_width, int window_height, 
								const Three_Vector& driver_position,
								double driver_field_of_view)
{
  const Three_Vector pos = m_position - driver_position;
  const double y_factor = 
	-1.0 / (pos [0] * tan (0.5 * deg_to_rad (driver_field_of_view)));
  const double aspect = double (window_width) / window_height;
  const double x_factor = -y_factor / aspect;

  const int x0 = to_pixels (window_width, x_factor, pos [1]) - 1;
  m_viewport.x = clip (x0, 0, window_width - 1);
  const int y0 = to_pixels (window_height, y_factor, pos [2]) - 1;
  m_viewport.y = clip (y0, 0, window_height - 1);

  const int x1 = to_pixels (window_width, x_factor, pos [1] - m_width);
  m_viewport.width = clip (x1, 0, window_width - 1) - m_viewport.x;
  const int y1 = to_pixels (window_height, y_factor, pos [2] + m_height);
  m_viewport.height = clip (y1, 0, window_height - 1) - m_viewport.y;
}


// Draw the mask.
void Vamos_Body::
Rear_View_Mirror::draw_mask_shape ()
{
  glStencilFunc (GL_ALWAYS, 1, 1);
  glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);

  mp_mask->activate ();

  glColor3d (1.0, 1.0, 1.0);
  glBegin (GL_QUADS);
  glTexCoord2d (0.0, 1.0);
  glVertex3d (m_position [0], m_position [1], m_position [2]);
  glTexCoord2d (1.0, 1.0);
  glVertex3d (m_position [0], m_position [1] - m_width, m_position [2]);
  glTexCoord2d (1.0, 0.0);
  glVertex3d (m_position [0], m_position [1] - m_width, 
			  m_position [2] + m_height);
  glTexCoord2d (0.0, 0.0);
  glVertex3d (m_position [0], m_position [1], m_position [2] + m_height);
  glEnd ();

  glFlush ();
}


// Use the pixels in the viewport to set the stencil buffer.
void Vamos_Body::
Rear_View_Mirror::set_stencil (int window_width, int window_height)
{
  unsigned char* stencil_buffer = make_stencil_buffer ();

  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();
  gluOrtho2D (0.0, double (window_width), 0.0, double (window_height));
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();
  glLoadIdentity ();
  glStencilFunc (GL_EQUAL, 1, 1);
  glStencilOp (GL_KEEP, GL_REPLACE, GL_REPLACE);
  glRasterPos2i (m_viewport.x, m_viewport.y);
  glDrawPixels (m_viewport.width, m_viewport.height, GL_STENCIL_INDEX,
  				GL_UNSIGNED_BYTE, stencil_buffer);
  glPopMatrix ();

  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();

  glFinish ();

  delete stencil_buffer;
}


// Grab the current buffer as an array of pixels.
unsigned char* Vamos_Body::
Rear_View_Mirror::make_stencil_buffer ()
{
  glReadBuffer (GL_BACK);
  const size_t elements = m_viewport.width * m_viewport.height;
  unsigned char* rgba_buffer = new unsigned char [4 * elements];

  glReadPixels (m_viewport.x, m_viewport.y, 
 				m_viewport.width, m_viewport.height, 
 				GL_RGBA, GL_UNSIGNED_BYTE, rgba_buffer);

  unsigned char* buffer = new unsigned char [elements];
  for (size_t i = 0; i < elements; i++)
	{
	  buffer [i] = rgba_buffer [4 * i];
	}
  delete rgba_buffer;
  return buffer;
}


void Vamos_Body::
Gl_Car::draw_rear_view (double aspect, int index)
{
  Rear_View_Mirror* mirror = m_mirrors [index];
  mirror->set_view ();
  view (mirror->get_direction (), mirror->get_center ());
}

// Set the dashboard.
void Vamos_Body::
Gl_Car::dashboard (Dashboard* dash)
{
  delete mp_dashboard;
  mp_dashboard = dash;
}

#include "3dmath.h"

void Vamos_Body::
Gl_Car::transform_body ()
{
  // Set the modelview matrix to be the identity matrix
  //glLoadIdentity ();
  
  // Translate and rotate the graphical representation of the object
  Three_Vector p = m_chassis.position ();
  //glTranslatef (p [0], p [1], p [2]);
  
  double angle;
  Three_Vector axis = m_chassis.axis_angle (&angle);
	
	
  //glRotatef (angle, axis [0], axis [1], axis [2]);
	
	
		// The rotation is performed about the origin of the body
		// system.  Since we want the effect to be a rotation about the
		// center of mass, we have to translate the object the distance
		// from the cm to the origin.
		Three_Vector cm = m_chassis.center_of_mass ();
		//glTranslatef (cm [0], cm [1], cm [2]);
  
	//update scenegraph nodes
  	if (cm_node.PtrValid())
	{
		QUATERNION goofyfoot;
		goofyfoot.Rotate(-3.141593*0.5, 0,0,1);
		
		VERTEX pos;
		pos.Set(p[0],p[1],p[2]);
		QUATERNION rot;
		rot.SetAxisAngle(angle*0.01745329, axis[0], axis[1], axis[2]);
		//rot.PostMultiply(goofyfoot);
		//rot.Normalize();
		//rot.DebugPrint();
		cm_node.GetRef().GetTransform().SetTranslation(pos);
		cm_node.GetRef().GetTransform().SetRotation(rot);
		VERTEX cmpos;
		cmpos.Set(cm[0],cm[1],cm[2]);
		origin_node.GetRef().GetTransform().SetTranslation(cmpos.ScaleR(-1.0));
		origin_node.GetRef().GetTransform().SetRotation(goofyfoot);
		
		if (collision_node.PtrValid())
		{
			double angle;
			Vamos_Geometry::Three_Vector aa = chassis().axis_angle(&angle);
			QUATERNION carorientation;
			//carorientation.SetAxisAngle(angle*0.01745329, aa[0], aa[1], aa[2]);
			carorientation.SetAxisAngle(Vamos_Geometry::deg_to_rad(angle), aa[0], aa[1], aa[2]);
			carorientation.Rotate(-3.141593/2.0, 1,0,0);
			
			//carorientation.Normalize();
			//carorientation.Rotate(3.141593*(3.0/2.0),0,1,0);
			//carorientation.Rotate(3.141593,0,0,1);
			//carorientation.SetAxisAngle(angle*0.01745329, aa[0], aa[2], aa[1]);
			
			VERTEX carposition = GetOriginNode()->TransformIntoWorldSpace();
			collision_node.GetRef().GetTransform().SetTranslation(carposition);
			collision_node.GetRef().GetTransform().SetRotation(carorientation);
		}
		
		/*cout << "Updated car position to ";pos.DebugPrint();
		//cout << "Updated CM position to ";cmpos.DebugPrint();
		cout << "World coords: ";origin_node.GetRef().GetWorldPosition().DebugPrint();
		cout << "Camera: ";game.cam.position.DebugPrint();*/
	}
	
	//transform the wheels
	int inum(0);
	VECTORITERATE(Wheel *, m_wheels, i)
	{
		(*i)->FenderTransform();
		
		//set tire squeal sound
		if (inum < 4)
		{
			float slide = (*i)->slide();
			float slidemultiple = 1.0; //for venzon tire code
			float squeal = slide*slidemultiple;
			if (squeal < 0)	//not sure this is necessary
				squeal = -squeal;
			float offset = 0.3;
			float gain = 0.5;
			squeal -= offset;
			squeal *= gain;
			if (squeal < 0)
				squeal = 0;
			if (squeal > 1)
				squeal = 1;
			bool snddebug = false;
			//snddebug = (inum != 0);
			float maxgain = 0.3;
			gain = squeal;
			if (gain > maxgain)
				gain = maxgain;
			if (gain < 0)
				gain = 0.0;
			if (!snddebug && tire_source[inum].PtrValid())
				tire_source[inum].GetRef().SetGainSmooth(gain,0.004);
			float pitch = 1.0-squeal;
			float pitchvariation = 0.4;
			pitch *= pitchvariation;
			pitch = pitch + (1.0-pitchvariation);
			if (pitch < 0.1)
				pitch = 0.1;
			if (pitch > 4.0)
				pitch = 4.0;
			if (!snddebug && tire_source[inum].PtrValid())
				tire_source[inum].GetRef().SetPitchSmooth(pitch,0.004);
			VERTEX wp((*i)->GetWorldPosition());
			if (!snddebug && tire_source[inum].PtrValid())
			{
				tire_source[inum].GetRef().SetPosition(wp);
				//cout << inum << " ";wp.DebugPrint();
			}
			
			//if (slide > 0) cout << slide << "," << squeal << endl;
			
			//add tire smoke
			float probmultiple = 100.0;
			float maxprob = 0.1;
			float prob = game.FrameTime() * probmultiple;
			prob = prob * squeal;
			if (prob < 0) //not sure this is necessary
				prob = -prob;
			VERTEX dir;
			dir.y = 1;
			game.particle.SetParams(1.0*slide, 2.0*slide, 5.0, 14.0,
					   0.3, 1.0, dir, 0.5, 1.0);
			if (prob > maxprob)
				prob = maxprob;
			game.particle.ProbAddParticle(wp, prob);
		}
		inum++;
	}
  
  	/*QUATERNION carrot;
	carrot.SetAxisAngle(angle*3.141593/180.0, axis[0], axis[2], axis[1]);
	QUATERNION goofyfoot;
	goofyfoot.Rotate(-3.141593/2.0, 1,0,0);
	double tempmat[16];
	goofyfoot.GetMat(tempmat);
	float lp[4];
	lp[0] = LightPosition[0];
	lp[1] = LightPosition[1];
	lp[2] = LightPosition[2];
	lp[3] = 0;
	VERTEX lpv;
	lpv.Set(lp);
	//goofyfoot.PostMultiply(carrot);
	lpv = goofyfoot.ReturnConjugate().RotateVec(lpv);
	lpv = carrot.ReturnConjugate().RotateVec(lpv);
	lp[0] = lpv.x;
	lp[1] = lpv.z;
	lp[2] = lpv.y;*/
	//glLightfv( GL_LIGHT1, GL_POSITION,  lp);
	
	//set brake illumination
	if (exterior_drawable.PtrValid())
		exterior_drawable.GetRef().SetSelfIllumination((brakesetting > 0));
}

#define SETFLOATINGILLUM(floatmodel) {if (floatmodel.Loaded()) floatmodel.SetTU(1, illum);}

// Render the car according to its current position and orientation.
void Vamos_Body::
Gl_Car::draw (bool transform, float opacity)
{
/*//  if (m_body_list_id != 0)
	{
		if (transform)
	  		transform_body ();

	  // Draw the body.
	  //glCallList (m_body_list_id);
		//joeexterior.Draw(0, 0);
		
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		//glDisable(GL_LIGHTING);
		//glDepthMask(0);
		
		bool illum = false;
		if (brakesetting > 0)
			illum = true;
		
		//game.utility.SelectTU(1);
		//glEnable(GL_TEXTURE_2D);
		//game.utility.SelectTU(0);

		
		glColor4f(1,1,1,opacity);
		//glColor4f(0.5,0,0,opacity);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glCullFace(GL_FRONT);
		glEnable(GL_CULL_FACE);
		//joecollision.DrawStatic();
		joeexterior.SetTU(1, illum);
		SETFLOATINGILLUM(joefloating_front);
		SETFLOATINGILLUM(joefloating_rear);
		joeexterior.DrawStatic();
		joeexterior.SetTU(1, false);
		
		glCullFace(GL_BACK);
		glColor4f(0,0,0,opacity);
		joeexterior.DrawStatic();
		
		glColor4f(1,1,1,opacity);
		glDisable(GL_CULL_FACE);
		
		joeinterior.DrawStatic();
		
		// Draw the wheels.
		//  std::for_each (m_wheels.begin (), m_wheels.end (), std::mem_fun (&Wheel::draw));
		VECTORITERATE(Wheel *, m_wheels, i)
		{
			JOEMODEL * floatingmodel = NULL;
			if ((*i)->frontback() == FRONT) floatingmodel = &joefloating_front;
			if ((*i)->frontback() == BACK) floatingmodel = &joefloating_rear;
			glPushMatrix();
			(*i)->FenderTransform();
			floatingmodel->DrawStatic();
			glPopMatrix();
			(*i)->draw();
		}
		
		//game.utility.SelectTU(1);
		//glEnable(GL_TEXTURE_2D);
		//game.utility.SelectTU(0);
		
		glEnable(GL_CULL_FACE);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glCullFace(GL_FRONT);
		glEnable(GL_CULL_FACE);
		glDepthMask(0);
		glColor4f(1,1,1,opacity);
		joeglass.SetTU(1, illum);
		joeglass.DrawStatic();
		//joeglass.DrawStatic();
		glDepthMask(1);

		glPopAttrib();
		
	}*/
}

void Vamos_Body::
Gl_Car::draw_interior ()
{
  //joeinterior.Draw();
}

void Vamos_Body::
Gl_Car::draw_dashboard ()
{
  mp_dashboard->set_tachometer (rad_s_to_rpm (engine ()->rotational_speed ()));
  mp_dashboard->set_speedometer (m_s_to_km_h (wheel (2)->speed ()));
  mp_dashboard->set_fuel_gauge (fuel_tank ()->fuel ());
  mp_dashboard->set_gear_indicator (transmission ()->gear ());
  mp_dashboard->set_steering_wheel (m_steer_key_control.value ());

  mp_dashboard->draw ();
  if (m_show_dashboard_extras)
	{
	  draw_dashboard_extras ();
	}
}

void Vamos_Body::
Gl_Car::draw_string (const std::string& str, double x, double y)
{
	game.font.Print(x, y, str.c_str(), 1, 5, 1);
  /*glRasterPos2d (x, y);
  for (std::string::const_iterator it = str.begin (); it != str.end (); it++)
	{
	  glutBitmapCharacter (GLUT_BITMAP_8_BY_13, *it);
	}*/
}

void Vamos_Body::
Gl_Car::draw_dashboard_extras ()
{ 
  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();

  // Set the dimensions of the drawing area.
  gluOrtho2D (0, 10, 0, 10);
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();
  glLoadIdentity ();
  
  glDisable (GL_DEPTH_TEST);
  glDisable (GL_LIGHTING);
  glDisable (GL_TEXTURE_2D);
  
  glColor3f (1.0, 1.0, 1.0);

  std::ostringstream b_stream;

  int rpm = 
 	int (Vamos_Geometry::rad_s_to_rpm (engine ()->rotational_speed ()));
  b_stream << "RPM " << rpm;
  draw_string (b_stream.str (), 0.4, 1.8);

  b_stream.str ("");
  b_stream << "Torque " << int (engine ()->drive_torque ()) << " Nm";
  draw_string (b_stream.str (), 0.4, 1.4);

  b_stream.str ("");
  b_stream << "Speed " << int (m_s_to_km_h (wheel (2)->speed ()))
		   << " km/h";
  draw_string (b_stream.str (), 0.4, 1.0);

  b_stream.str ("");
  b_stream << "Mass " << int (m_chassis.mass ()) << " kg";
  draw_string (b_stream.str (), 0.4, 0.6);

  b_stream.str ("");
  char gear = 'N';
  if (transmission ()->gear () == -1)
	{
	  gear = 'R';
	}
  else if (transmission ()->gear () > 0)
	{
	  gear = transmission ()->gear () + '0';
	}
  b_stream << "Gear " << gear;
  draw_string (b_stream.str (), 0.4, 0.2);

  b_stream.str ("");
  b_stream << "Fuel " << std::setprecision (1)
		   << mp_fuel_tank->fuel () << " L";
  draw_string (b_stream.str (), 2.8, 0.2);


  b_stream.str ("");
  b_stream << "Slip Ratios";
  draw_string (b_stream.str (), 2.8, 1.4);

  b_stream.str ("");
  b_stream.setf (std::ios::fixed);
  b_stream << std::setprecision (3) << m_wheels [0]->slip ();
  draw_string (b_stream.str (), 2.8, 1.0);

  b_stream.str ("");
  b_stream << std::setprecision (3) << m_wheels [1]->slip ();
  draw_string (b_stream.str (), 3.6, 1.0);

  b_stream.str ("");
  b_stream << std::setprecision (3) << m_wheels [2]->slip ();
  draw_string (b_stream.str (), 2.8, 0.6);

  b_stream.str ("");
  b_stream << std::setprecision (3) << m_wheels [3]->slip ();
  draw_string (b_stream.str (), 3.6, 0.6);

  glEnable (GL_DEPTH_TEST);
  glEnable (GL_LIGHTING);
  glEnable (GL_TEXTURE_2D);

  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();
  glMatrixMode (GL_MODELVIEW);
  glPopMatrix ();
}

// Perform the transformations for the driver's view.
void Vamos_Body::
Gl_Car::view (double pan, const Three_Vector& view_position)
{
  if (pan == 0.0)
	{
	  pan = m_pan_key_control.value ();
	}

  // Find the angle-axis representation of the car's orientation.
  double angle;
  Three_Vector axis = m_chassis.axis_angle (&angle);

  // Rotate the view.
  glRotated (90, 0.0, 1.0, 0.0);
  glRotated (-90, 1.0, 0.0, 0.0);
  glRotated (-angle, axis [0], axis [1], axis [2]);

  Three_Vector z = m_chassis.rotate_out (Three_Vector (0.0, 0.0, 1.0));
  glRotated (-pan, z [0], z [1], z [2]);

  Three_Vector pos = 
    -m_chassis.transform_out (view_position - m_chassis.center_of_mass ());
  glTranslated (pos [0], pos [1], pos [2]);
}

//extern string car_file;

void Vamos_Body::
Gl_Car::engine_sound (map <float, string> & pitch_soundmap)
{
	if (!game.sound.Enabled())
		return;
	
	if (pitch_soundmap.empty())
	{
		SOUNDSOURCE & newsource = game.sound.NewSource("cars/" + m_car_file + "/engine");
		//SOUNDSOURCE & newsource = game.sound.NewSource("test/oggtest");
		//real_engine_sample.SetPtr(newsource);
		engine_samples[7000].SetPtr(newsource);
		//cout << "Allocating sound: " << "cars/" + m_car_file + "/engine" << ", " << &newsource << endl;
		newsource.SetGain(0);
		newsource.SetLoop(true);
		newsource.Set3DEffects(true);
		//newsource.AddFilter().SetFilterOrder1(0.1, 0.0, 0.0, 0.9); //simple low-pass filter
		newsource.Play();
	}
	else
	{
		for (map <float, string>::iterator i = pitch_soundmap.begin(); i != pitch_soundmap.end(); i++)
		{
			SOUNDSOURCE & newsource = game.sound.NewSource("cars/" + m_car_file + "/sounds/" + i->second);
			//SOUNDSOURCE & newsource = game.sound.NewSource("test/oggtest");
			engine_samples[i->first].SetPtr(newsource);
			//cout << "Allocating sound: " << "cars/" + m_car_file + "/engine" << ", " << &newsource << endl;
			newsource.SetGain(0);
			newsource.SetLoop(true);
			newsource.Set3DEffects(true);
			//newsource.AddFilter().SetFilterOrder1(0.1, 0.0, 0.0, 0.9); //simple low-pass filter
			newsource.Play();
		}
	}

	for (int i = 0; i < 4; i++)
	{
		SOUNDSOURCE & newsource = game.sound.NewSource("sounds/tire_squeal");
		int samples = newsource.GetSoundBuffer().GetSoundInfo().GetSamples();
		newsource.SeekToSample((samples/4)*i);
		tire_source[i].SetPtr(newsource);
		newsource.SetGain(0);
		newsource.SetLoop(true);
		newsource.Set3DEffects(true);
		newsource.Play();
	}
}

void Vamos_Body::
Gl_Car::SetSoundPositionAndVelocity(VERTEX & spos, VERTEX & svel)
{
	for (map <float, BIPOINTER <SOUNDSOURCE> >::iterator i = engine_samples.begin(); i != engine_samples.end(); i++)
	{
		i->second.GetPtr()->SetPosition(spos);
		i->second.GetPtr()->SetVelocity(svel);
	}
}

void Vamos_Body::
Gl_Car::SetPitchAndGainSmoothly(float erpm, float gain, float fade)
{
	SOUNDSOURCE * interp1(NULL);
	float rpm1(0);
	SOUNDSOURCE * interp2(NULL);
	float rpm2(100000);
	
	for (map <float, BIPOINTER <SOUNDSOURCE> >::iterator i = engine_samples.begin(); i != engine_samples.end(); i++)
	{
		if (!interp1 || (i->first < erpm && i->first > rpm1)) //find the sample for the RPM just below the current RPM
		{
			interp1 = i->second.GetPtr();
			rpm1 = i->first;
		}
		if (i->first > erpm && i->first < rpm2) //find the sample for the RPM just above the current RPM
		{
			interp2 = i->second.GetPtr();
			rpm2 = i->first;
		}
	}
	
	//cout << erpm << "," << rpm1 << "(" << interp1 << ")--" << rpm2 << "(" << interp2 << ")" << endl;
	if (interp2 == interp1)
		interp2 = NULL;
	
	float crossfadegain = 0.0;
	if (interp2)
	{
		crossfadegain = (erpm - rpm1)/(rpm2 - rpm1);
	}
	
	//fade *= 10.0;
	
	if (interp1)
	{
		interp1->SetPitchSmooth(erpm/rpm1, fade);
		//interp1->SetGainSmooth(gain*(1.0-crossfadegain), fade);
		
		//interp1->SetPitch(erpm/rpm1);
		interp1->SetGain(gain*(1.0-crossfadegain));
	}
	
	if (interp2)
	{
		interp2->SetPitchSmooth(erpm/rpm2, fade);
		//interp2->SetGainSmooth(gain*crossfadegain, fade);
		
		//interp2->SetPitch(erpm/rpm2);
		interp2->SetGain(gain*crossfadegain);
	}
	
	//quiet the other sounds
	for (map <float, BIPOINTER <SOUNDSOURCE> >::iterator i = engine_samples.begin(); i != engine_samples.end(); i++)
	{
		if (i->second.GetPtr() != interp1 && i->second.GetPtr() != interp2)
			i->second.GetPtr()->SetGain(0);
			//i->second.GetPtr()->SetGainSmooth(0,fade);
	}
}

double Vamos_Body::
Gl_Car::engine_pitch ()
{
  return engine ()->rotational_speed ();
}

double Vamos_Body::
Gl_Car::engine_volume ()
{
  return 1.0 + m_throttle_volume_factor * engine ()->throttle ()
	+ m_engine_speed_volume_factor * engine ()->rotational_speed ();
}

/*SOUNDSOURCE * Vamos_Body::Gl_Car::GetSoundSource()
{
	if (real_engine_sample.PtrValid())
		return real_engine_sample.GetPtr();
	else
		return NULL;
}

SOUNDSOURCE * Vamos_Body::Gl_Car::GetTireSoundSource(int i)
{
	if (tire_source[i].PtrValid())
		return tire_source[i].GetPtr();
	else
		return NULL;
}*/

void Vamos_Body::Gl_Car::SetPaint(int pid)
{
	if (num_paintjobs <= 0)
		return;
	
	char texfile[1024];
	if (pid < 0)
		pid = num_paintjobs - ((-pid) % num_paintjobs);
	pid = pid % num_paintjobs;
	sprintf(texfile, "cars/%s/textures/body%02i.png", m_car_file.c_str(), pid);
	//joeexterior.Texture(game.settings.GetFullDataPath(texfile), 0);
	if (exterior_drawable.PtrValid())
	{
		TEXTUREINFO texinfo;
		texinfo.SetName(texfile);
		exterior_drawable.GetRef().SetDiffuseMap(texinfo);
		TEXTUREINFO misctexinfo;
		if (game.utility.FileExists(game.settings.GetFullDataPath("cars/"+m_car_file+"/textures/body-misc1.png")))
		{
			misctexinfo.SetName("cars/"+m_car_file+"/textures/body-misc1.png");
			exterior_drawable.GetRef().SetMiscMap1(misctexinfo);
		}
		else
		{
			stringstream msg;
			msg << "Car " << m_car_file << " has no gloss map; using a default contant shiny/reflective texture" << endl;
			game.WriteDebuggingData(msg.str());
			
			misctexinfo.SetName("textures/effects/white.png");
			exterior_drawable.GetRef().SetMiscMap1(misctexinfo);
		}
		
		if (game.utility.FileExists(game.settings.GetFullDataPath("cars/"+m_car_file+"/textures/brake.png")))
		{
			TEXTUREINFO texinfo2;
			texinfo2.SetName("cars/"+m_car_file+"/textures/brake.png");
			exterior_drawable.GetRef().SetAdditiveMap1(texinfo2);
		}
		
		//inform wheels of body paint
		VECTORITERATE(Wheel *, m_wheels, i)
		{
			(*i)->set_paint(&texinfo,&misctexinfo);
		}
		//cout << "Diffuse map set to " << texfile << endl;
	}
	
	cur_paintjob = pid;
}

void Vamos_Body::Gl_Car::SetReflectionTexture(TEXTURE_HANDLE * reftid, bool cubemapping)
{
	/*int targetid = 2;
	
	if (cubemapping)
	{
		joeexterior.StaticCubeReflectionTextureID(reftid, targetid);
		joeglass.StaticCubeReflectionTextureID(reftid, targetid);
	
		if (joefloating_front.Loaded()) joefloating_front.StaticCubeReflectionTextureID(reftid, targetid);
		if (joefloating_rear.Loaded()) joefloating_rear.StaticCubeReflectionTextureID(reftid, targetid);
	}
	else
	{
		joeexterior.ReflectionTextureID(reftid, targetid);
		joeglass.ReflectionTextureID(reftid, targetid);
	
		if (joefloating_front.Loaded()) joefloating_front.ReflectionTextureID(reftid, targetid);
		if (joefloating_rear.Loaded()) joefloating_rear.ReflectionTextureID(reftid, targetid);
	}*/
}

void Vamos_Body::Gl_Car::SetDynamicCubeReflectionTexture(GLuint reftid)
{
	/*int targetid = 2;
	
	joeexterior.DynamicCubeReflectionTextureID(reftid, targetid);
	joeglass.DynamicCubeReflectionTextureID(reftid, targetid);

	if (joefloating_front.Loaded()) joefloating_front.DynamicCubeReflectionTextureID(reftid, targetid);
	if (joefloating_rear.Loaded()) joefloating_rear.DynamicCubeReflectionTextureID(reftid, targetid);*/
}
